@@ -1,9 +1,11 @@
 use crate::{
-    db::OffsetHandler,
-    error::{Err, Error},
+    error::Error,
     map::MappedFile,
-    offsets::{offset_client::OffsetClient, offset_server::Offset, OffsetsRequest},
+    offsets::{
+        offset_client::OffsetClient, offset_server::Offset, OffsetsRequest, OffsetsResponse,
+    },
     primitives::{protected_write, rtl_rb_tree_insert, ProtectionGuard, RtlMutex},
+    server::OffsetHandler,
     util::to_wide,
 };
 use log::debug;
@@ -25,6 +27,7 @@ use std::{
     ptr,
     ptr::null_mut,
 };
+use tonic::transport::{Certificate, Channel, ClientTlsConfig};
 use winapi::{
     shared::{
         guiddef::GUID,
@@ -34,6 +37,7 @@ use winapi::{
             UNICODE_STRING,
         },
         ntstatus::STATUS_SUCCESS,
+        winerror::ERROR_FILE_NOT_FOUND,
     },
     um::{
         libloaderapi::{GetModuleHandleW, GetProcAddress, LoadLibraryA},
@@ -51,7 +55,8 @@ use winapi::{
     },
 };
 
-/// The context of Nt functions and statics
+/// The context of internal Nt functions and statics that are
+/// required for the mapper to work
 #[allow(non_snake_case)]
 pub struct NtContext<'a> {
     LdrpHashTable: RtlMutex<'a, [LIST_ENTRY; 32]>,
@@ -71,7 +76,7 @@ impl<'a> NtContext<'a> {
         unsafe {
             let dos_header = ntdll as *const IMAGE_DOS_HEADER;
             if dos_header.is_null() {
-                return Err(Error(Err::FileNotFound));
+                return Err(Error::NtDllNotLoaded);
             }
             let nt_headers = &*((dos_header as *const u8).offset((*dos_header).e_lfanew as isize)
                 as *const IMAGE_NT_HEADERS64);
@@ -80,7 +85,7 @@ impl<'a> NtContext<'a> {
                     .VirtualAddress as isize,
             ) as *const IMAGE_DEBUG_DIRECTORY);
             if debug_entry.Type != IMAGE_DEBUG_TYPE_CODEVIEW {
-                return Err(Error(Err::NtDllDebugType));
+                return Err(Error::NtDllDebugType);
             }
             let codeview_entry = &*((dos_header as *const u8)
                 .offset((*debug_entry).AddressOfRawData as isize)
@@ -89,7 +94,7 @@ impl<'a> NtContext<'a> {
                 .rsds_signature
                 .eq_ignore_ascii_case("RSDS".as_bytes())
             {
-                return Err(Error(Err::NtDllRsdsSig));
+                return Err(Error::NtDllRsdsSig);
             }
             Ok(format!(
                 "{:08X}{:04X}{:04X}{}{:X}",
@@ -108,22 +113,14 @@ impl<'a> NtContext<'a> {
         }
     }
 
-    /// Resolves the context used by the mapper
+    /// Processes an offsets response
     ///
     /// # Arguments
     ///
-    /// `server_hostname`: The hostname of the endpoint of the PDB server
-    /// `server_port`: The port of the endpoint of the PDB server
-    pub async fn resolve(server_hostname: &str, server_port: u16) -> anyhow::Result<NtContext<'a>> {
-        let mut client =
-            OffsetClient::connect(format!("http://{}:{}", server_hostname, server_port)).await?;
-        let ntdll = unsafe { GetModuleHandleW(to_wide("ntdll").as_ptr()) as *const u8 };
-        let request = tonic::Request::new(OffsetsRequest {
-            ntdll_hash: NtContext::get_ntdll_hash(ntdll)?,
-        });
-        let response = client.get_offsets(request).await?.into_inner();
+    /// `response`: The response to process
+    fn process_response(response: OffsetsResponse, ntdll: *const u8) -> NtContext<'a> {
         unsafe {
-            Ok(NtContext {
+            NtContext {
                 LdrpHashTable: RtlMutex::from_ref(
                     &mut *(ntdll.offset(response.ldrp_hash_table as isize)
                         as *mut [LIST_ENTRY; 32]),
@@ -148,10 +145,78 @@ impl<'a> NtContext<'a> {
                     &mut *(ntdll.offset(response.ldrp_module_datatable_lock as isize)
                         as *mut RTL_SRWLOCK),
                 ),
-            })
+            }
         }
     }
 
+    /// Resolves the context used by the mapper
+    ///
+    /// # Arguments
+    ///
+    /// `server_hostname`: The hostname of the endpoint of the PDB server
+    ///
+    /// `server_port`: The port of the endpoint of the PDB server
+    pub async fn resolve<S: AsRef<str>>(
+        server_hostname: S,
+        server_port: u16,
+    ) -> anyhow::Result<NtContext<'a>> {
+        let channel = Channel::from_shared(format!(
+            "http://{}:{}",
+            server_hostname.as_ref(),
+            server_port
+        ))?;
+        let mut client = OffsetClient::new(channel.connect().await?);
+        let ntdll = unsafe { GetModuleHandleW(to_wide("ntdll").as_ptr()) as *const u8 };
+        let request = tonic::Request::new(OffsetsRequest {
+            ntdll_hash: NtContext::get_ntdll_hash(ntdll)?,
+        });
+        let response = client.get_offsets(request).await?.into_inner();
+        Ok(NtContext::process_response(response, ntdll))
+    }
+
+    /// Resolves the context used by the mapper, over a secure TLS connection
+    ///
+    /// # Arguments
+    ///
+    /// `server_hostname`: The hostname of the endpoint of the PDB server
+    ///
+    /// `server_port`: The port of the endpoint of the PDB server
+    ///
+    /// `ca_cert`: A custom CA certificate. This is necessary if you have a
+    /// self-signed certificate on the other end. If not specified, webPKI
+    /// will use their store to verify the endpoint
+    ///
+    /// `domain`: The domain name to be verified
+    pub async fn resolve_tls<S: AsRef<str>>(
+        server_hostname: S,
+        server_port: u16,
+        ca_cert: Option<Certificate>,
+        domain: Option<S>,
+    ) -> anyhow::Result<NtContext<'a>> {
+        let mut tls_config = ClientTlsConfig::new();
+        // add a CA certificate in case a self-signed cert is used
+        if let Some(ca_cert) = ca_cert {
+            tls_config = tls_config.ca_certificate(ca_cert);
+        }
+        // add a pinned domain in case domain name verification is wanted
+        if let Some(domain) = domain {
+            tls_config = tls_config.domain_name(domain.as_ref());
+        }
+        let channel = Channel::from_shared(format!(
+            "http://{}:{}",
+            server_hostname.as_ref(),
+            server_port
+        ))?
+        .tls_config(tls_config)?;
+        let mut client = OffsetClient::new(channel.connect().await?);
+        let ntdll = unsafe { GetModuleHandleW(to_wide("ntdll").as_ptr()) as *const u8 };
+        let request = tonic::Request::new(OffsetsRequest {
+            ntdll_hash: NtContext::get_ntdll_hash(ntdll)?,
+        });
+        let response = client.get_offsets(request).await?.into_inner();
+        Ok(NtContext::process_response(response, ntdll))
+    }
+
     /// Resolves the context used by the mapper
     ///
     /// # Arguments
@@ -163,34 +228,7 @@ impl<'a> NtContext<'a> {
             ntdll_hash: NtContext::get_ntdll_hash(ntdll)?,
         });
         let response = handler.get_offsets(request).await?.into_inner();
-        unsafe {
-            Ok(NtContext {
-                LdrpHashTable: RtlMutex::from_ref(
-                    &mut *(ntdll.offset(response.ldrp_hash_table as isize)
-                        as *mut [LIST_ENTRY; 32]),
-                    &mut *(ntdll.offset(response.ldrp_module_datatable_lock as isize)
-                        as *mut RTL_SRWLOCK),
-                ),
-                LdrpHandleTlsData: std::mem::transmute(
-                    ntdll.offset(response.ldrp_handle_tls_data as isize),
-                ),
-                LdrpReleaseTlsEntry: std::mem::transmute(
-                    ntdll.offset(response.ldrp_release_tls_entry as isize),
-                ),
-                LdrpMappingInfoIndex: RtlMutex::from_ref(
-                    &mut *(ntdll.offset(response.ldrp_mapping_info_index as isize)
-                        as *mut RTL_RB_TREE),
-                    &mut *(ntdll.offset(response.ldrp_module_datatable_lock as isize)
-                        as *mut RTL_SRWLOCK),
-                ),
-                LdrpModuleBaseAddressIndex: RtlMutex::from_ref(
-                    &mut *(ntdll.offset(response.ldrp_module_base_address_index as isize)
-                        as *mut RTL_RB_TREE),
-                    &mut *(ntdll.offset(response.ldrp_module_datatable_lock as isize)
-                        as *mut RTL_SRWLOCK),
-                ),
-            })
-        }
+        Ok(NtContext::process_response(response, ntdll))
     }
 }
 
@@ -213,13 +251,14 @@ struct IMAGE_DEBUG_CODEVIEW {
 /// # Arguments
 ///
 /// `image_base`: The base address of the image
-pub unsafe fn set_as_primary_image(image_base: PVOID) -> PVOID {
+unsafe fn set_as_primary_image(image_base: PVOID) -> PVOID {
     let peb = NtCurrentPeb();
     let old_base = (*peb).ImageBaseAddress;
     (*peb).ImageBaseAddress = image_base;
     old_base
 }
 
+/// A portable executable that is all nicely wrapped up in a class
 pub struct PortableExecutable<'a> {
     file: MappedFile,
     file_name: Vec<u16>,
@@ -349,7 +388,7 @@ impl<'a> PortableExecutable<'a> {
                 self.file.contents() as u64,
             ) == false as u8
         } {
-            return Err(Error(Err::ExceptionTableEntry));
+            return Err(Error::ExceptionTableEntry);
         }
         Ok(())
     }
@@ -363,7 +402,7 @@ impl<'a> PortableExecutable<'a> {
         }
         // remove the table from the process
         if unsafe { RtlDeleteFunctionTable(exception_table.as_mut_ptr()) == false as u8 } {
-            return Err(Error(Err::ExceptionTableEntry));
+            return Err(Error::ExceptionTableEntry);
         }
         Ok(())
     }
@@ -383,13 +422,13 @@ impl<'a> PortableExecutable<'a> {
         let tls_directory = self
             .file
             .get_rva::<IMAGE_TLS_DIRECTORY>(tls_directory.VirtualAddress as isize)
-            .ok_or(Error(Err::TLSOutOfBounds))?;
+            .ok_or(Error::TLSOutOfBounds)?;
         let mut tls_callback = unsafe {
             self.file
                 .get_rva::<Option<unsafe extern "stdcall" fn(PVOID, u32, PVOID)>>(
                     ((*tls_directory).AddressOfCallBacks - self.file.contents() as u64) as isize,
                 )
-                .ok_or(Error(Err::CallbackOutOfBounds))?
+                .ok_or(Error::CallbackOutOfBounds)?
         };
         // execute all of the TLS callbacks
         unsafe {
@@ -408,7 +447,7 @@ impl<'a> PortableExecutable<'a> {
         if unsafe { (self.context.LdrpReleaseTlsEntry)(self.loader_entry.as_mut().get_mut(), 0) }
             != STATUS_SUCCESS
         {
-            Err(Error(Err::TLSData))
+            Err(Error::TLSData)
         } else {
             Ok(())
         }
@@ -419,7 +458,7 @@ impl<'a> PortableExecutable<'a> {
         if unsafe { (self.context.LdrpHandleTlsData)(self.loader_entry.as_mut().get_mut()) }
             != STATUS_SUCCESS
         {
-            Err(Error(Err::TLSData))
+            Err(Error::TLSData)
         } else {
             Ok(())
         }
@@ -436,7 +475,7 @@ impl<'a> PortableExecutable<'a> {
         let exception_table = self
             .file
             .get_rva_mut::<IMAGE_RUNTIME_FUNCTION_ENTRY>(exception_dir.VirtualAddress as isize)
-            .ok_or(Error(Err::ExceptionTableOutOfBounds))?;
+            .ok_or(Error::ExceptionTableOutOfBounds)?;
         return Ok(unsafe {
             std::slice::from_raw_parts_mut(
                 exception_table,
@@ -463,7 +502,7 @@ impl<'a> PortableExecutable<'a> {
                 &mut hash,
             ) != STATUS_SUCCESS
             {
-                return Err(Error(Err::LdrEntry));
+                return Err(Error::LdrEntry);
             }
         }
         self.loader_entry.BaseNameHashValue = hash;
@@ -515,7 +554,9 @@ impl<'a> PortableExecutable<'a> {
         // first make sure we got all of the required functions
         let mut file = MappedFile::load(path)?;
         let path = Path::new(path);
-        let file_name = path.file_name().ok_or(Error(Err::FileNotFound))?;
+        let file_name = path
+            .file_name()
+            .ok_or_else(|| std::io::Error::from_raw_os_error(ERROR_FILE_NOT_FOUND as i32))?;
         let file_path = path.canonicalize()?;
         debug!(
             "Loading {} at path {}",
@@ -615,7 +656,7 @@ impl<'a> PortableExecutable<'a> {
             // we transmute here because I have no earthly idea how to return a generic function
             let entry_point: PLDR_INIT_ROUTINE = std::mem::transmute(
                 file.get_rva::<u8>(address_of_ep as isize)
-                    .ok_or(Error(Err::EPOutOfBounds))?,
+                    .ok_or(Error::EPOutOfBounds)?,
             );
             Ok(entry_point)
         }
@@ -632,15 +673,15 @@ impl<'a> PortableExecutable<'a> {
         unsafe {
             let dos_header = &*file
                 .get_rva::<IMAGE_DOS_HEADER>(0)
-                .ok_or(Error(Err::DOSOutOfBounds))?;
+                .ok_or(Error::DOSOutOfBounds)?;
             let nt_headers = &*file
                 .get_rva::<IMAGE_NT_HEADERS64>(dos_header.e_lfanew as isize)
-                .ok_or(Error(Err::NTOutOfBounds))?;
+                .ok_or(Error::NTOutOfBounds)?;
             // ensure supported architecture
             if nt_headers.FileHeader.Machine != IMAGE_FILE_MACHINE_AMD64
                 || nt_headers.OptionalHeader.Magic != IMAGE_NT_OPTIONAL_HDR64_MAGIC
             {
-                return Err(Error(Err::UnsupportedArch));
+                return Err(Error::UnsupportedArch);
             }
             // load section headers
             let section_headers = std::slice::from_raw_parts(
@@ -650,7 +691,7 @@ impl<'a> PortableExecutable<'a> {
                     nt_headers.FileHeader.NumberOfSections as usize
                         * std::mem::size_of::<IMAGE_SECTION_HEADER>(),
                 )
-                .ok_or(Error(Err::SectOutOfBounds))?,
+                .ok_or(Error::SectOutOfBounds)?,
                 nt_headers.FileHeader.NumberOfSections as usize,
             );
             Ok((nt_headers, section_headers))
@@ -669,7 +710,7 @@ impl<'a> PortableExecutable<'a> {
                 self.section_protections.push(ProtectionGuard::new(
                     self.file
                         .get_rva_mut::<c_void>(section.VirtualAddress as isize)
-                        .ok_or(Error(Err::SectOutOfBounds))?,
+                        .ok_or(Error::SectOutOfBounds)?,
                     *section.Misc.VirtualSize() as usize,
                     PortableExecutable::section_flags_to_prot(section.Characteristics),
                 )?)
@@ -694,7 +735,7 @@ impl<'a> PortableExecutable<'a> {
                     iat_directory.VirtualAddress as isize,
                     iat_directory.Size as usize,
                 )
-                .ok_or(Error(Err::IDOutOfBounds))?;
+                .ok_or(Error::IDOutOfBounds)?;
             self.resolve_import_descriptors(std::slice::from_raw_parts(
                 iat_entry,
                 (iat_directory.Size as usize) / std::mem::size_of::<IMAGE_IMPORT_DESCRIPTOR>(),
@@ -724,7 +765,7 @@ impl<'a> PortableExecutable<'a> {
             let name = self
                 .file
                 .get_rva::<i8>(entry.Name as isize)
-                .ok_or(Error(Err::LibNameOutOfBounds))?;
+                .ok_or(Error::LibNameOutOfBounds)?;
             unsafe {
                 debug!("Loading library {:?}", CStr::from_ptr(name));
                 let library = LoadLibraryA(name);
@@ -738,7 +779,7 @@ impl<'a> PortableExecutable<'a> {
                 let mut thunk = self
                     .file
                     .get_rva_mut::<IMAGE_THUNK_DATA>(entry.FirstThunk as isize)
-                    .ok_or(Error(Err::IATOutOfBounds))?;
+                    .ok_or(Error::IATOutOfBounds)?;
                 while (*thunk).u1.AddressOfData() != &0 {
                     let proc_name = if ((*thunk).u1.Ordinal() & IMAGE_ORDINAL_FLAG) != 0 {
                         (*(*thunk).u1.Ordinal() & !IMAGE_ORDINAL_FLAG) as *const i8
@@ -746,7 +787,7 @@ impl<'a> PortableExecutable<'a> {
                         &(*self
                             .file
                             .get_rva::<IMAGE_IMPORT_BY_NAME>(*(*thunk).u1.AddressOfData() as isize)
-                            .ok_or(Error(Err::ProcNameOutOfBounds))?)
+                            .ok_or(Error::ProcNameOutOfBounds)?)
                         .Name as *const i8
                     };
                     if ((*thunk).u1.Ordinal() & IMAGE_ORDINAL_FLAG) != 0 {
@@ -757,7 +798,7 @@ impl<'a> PortableExecutable<'a> {
                     } else {
                         // if it is a pointer, make sure it is not null
                         if proc_name.is_null() {
-                            return Err(Error(Err::NullProcName).into());
+                            return Err(Error::NullProcName.into());
                         }
                         debug!(
                             "Loading procedure with name {:?}",
@@ -851,12 +892,13 @@ mod test {
     use lazy_static::lazy_static;
     use serial_test::serial;
     use std::{
-        net::{IpAddr, Ipv4Addr, SocketAddr},
+        net::{IpAddr, Ipv4Addr, SocketAddr, SocketAddrV4},
         ptr::null,
         sync::Once,
         thread,
     };
     use tokio::runtime::Runtime;
+    use tonic::transport::Identity;
     use winapi::{
         shared::winerror::{ERROR_BAD_EXE_FORMAT, ERROR_MOD_NOT_FOUND, ERROR_PROC_NOT_FOUND},
         um::libloaderapi::GetModuleFileNameW,
@@ -875,7 +917,8 @@ mod test {
         INIT.call_once(|| {
             let server = Server::new(
                 SocketAddr::new(IpAddr::V4(Ipv4Addr::new(127, 0, 0, 1)), 42221),
-                "test/cache.json".into(),
+                "test/cache.json",
+                None,
             )
             .unwrap();
             thread::spawn(move || {
@@ -917,7 +960,7 @@ mod test {
             .unwrap()
             .downcast()
             .unwrap();
-        assert_eq!(err.0, Err::EPOutOfBounds);
+        assert_eq!(err, Error::EPOutOfBounds);
     }
 
     #[test]
@@ -954,7 +997,7 @@ mod test {
             .unwrap()
             .downcast()
             .unwrap();
-        assert_eq!(err.0, Err::UnsupportedArch);
+        assert_eq!(err, Error::UnsupportedArch);
     }
 
     #[test]
@@ -1022,4 +1065,23 @@ mod test {
         }
         assert_eq!(original_handle, unsafe { GetModuleHandleW(null()) });
     }
+
+    #[tokio::test]
+    #[serial]
+    async fn tls_server() {
+        // start the server
+        let endpoint = SocketAddr::V4(SocketAddrV4::new(Ipv4Addr::new(127, 0, 0, 1), 44443));
+        // we stole these certificates from the tonic test suite
+        let cert = tokio::fs::read("test/server.pem").await.unwrap();
+        let key = tokio::fs::read("test/server.key").await.unwrap();
+        let identity = Identity::from_pem(cert, key);
+        let server = Server::new(endpoint, "test/tls_cache.json", Some(identity)).unwrap();
+        // resolve
+        let ca_cert = tokio::fs::read("test/ca.pem").await.unwrap();
+        let ca_cert = Certificate::from_pem(ca_cert);
+        tokio::select! {
+            _ = server.run() => {},
+            res = NtContext::resolve_tls("localhost", 44443, Some(ca_cert), None) => { res.unwrap(); }
+        };
+    }
 }
